/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellML.h"
#include <nspr.h>
#include <nsArray.h>
#include <nsString.h>
#include <nsWeakPtr.h>
#include <nsDeque.h>
#include <necko/nsIServerSocket.h>
#include "mozCellMLIIntegrator.h"
#include "mozCellMLIVariable.h"
#include <nsIEventQueue.h>
#include <nsIAsyncOutputStream.h>
#include <nsIAsyncInputStream.h>
#include <necko/nsISocketTransport.h>
#include "IntegratorCommand.h"

class mozCellMLVariable;

class mozCellMLIntegrator
  : public mozCellMLIIntegrator,
    public nsIInputStreamCallback,
    public nsIOutputStreamCallback,
    public nsIServerSocketListener
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_MOZCELLMLIINTEGRATOR
  NS_DECL_NSIINPUTSTREAMCALLBACK
  NS_DECL_NSIOUTPUTSTREAMCALLBACK
  NS_DECL_NSISERVERSOCKETLISTENER

  mozCellMLIntegrator(nsIServerSocket* listener,
                      nsCOMArray<mozCellMLVariable>& idifferentials,
                      nsCOMArray<mozCellMLVariable>& isimples,
                      nsACString& filename,
                      PRUint32 ihidden_higher);

private:
  nsCOMPtr<nsIServerSocket> listener;
  nsCOMPtr<nsISocketTransport> transport;
  nsCOMPtr<nsIAsyncInputStream> input;
  nsCOMPtr<nsIAsyncOutputStream> output;
  nsCOMPtr<nsIEventQueue> eventQueue;
  // This array holds the owning references...
  nsCOMArray<mozCellMLVariable> differentials, simples;
  // This one shares the owning reference with the nsCOMArray...
  mozCellMLIVariable** variable;
  PRUint32 numvars;
  nsCOMArray<nsIWeakReference> next_run_watchers;
  char* in_buffer;
  nsCString out_buffer;
  
  PRUint32 max_packet, buf_pos, hidden_higher;
  PRBool awaiting_callback;

  struct IntegrationRun
  {
    IntegrationRun(nsCOMArray<nsIWeakReference>& x) : watchers(x) {}
    nsCOMArray<nsIWeakReference> watchers;
  };

  NS_IMETHOD ProcessPacket(IntegratorReply reply, void* data, PRUint32 length);
  
  nsDeque IntegrationRuns;

  PRUint32 wantcount;
  nsCString filename;

  NS_IMETHOD Enqueue(const void* ptr, PRUint32 length);

  ~mozCellMLIntegrator();
};
