function InsertVariable(box, name, type, initialValue, units)
{
  var li = document.createElement("listitem");
  li.setAttribute("id", name);
  var lc = document.createElement("listcell");
  lc.setAttribute("label", name);
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("label", type);
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("label", initialValue);
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("label", units);
  li.appendChild(lc);
  box.appendChild(li);
}

function ChangeInitialValue()
{
  var tf = document.getElementById("fieldvalue");
  var value = parseFloat(tf.value);
  if (isNaN(value))
    return;
  var tree = document.getElementById('variable-tree');
  var idx = tree.selectedIndex;
  if (idx == -1)
    return;
  if ((window.variables[idx].type ==
       Components.interfaces.mozCellMLIVariable.BOUND) ||
       (window.variables[idx].type ==
        Components.interfaces.mozCellMLIVariable.SIMPLE))
    return;

  var listitem = tree.selectedItem;
  var valueSubs = listitem.childNodes;
  valueSubs.item(2).setAttribute("label", value);
  // Now we have to change the model...
  window.variables[idx].initialValue = value;
  return value;
}

function SelectVariable()
{
  var tree = document.getElementById('variable-tree');
  var idx = tree.selectedIndex;
  if (idx == -1)
    return;
  var listitem = tree.selectedItem;
  var valueSubs = listitem.childNodes;
  var value = valueSubs.item(2).getAttribute("label");
  var name = valueSubs.item(0).getAttribute("label");
  var tf = document.getElementById("fieldvalue");
  tf.readonly = (window.variables[idx].type ==
                 Components.interfaces.mozCellMLIVariable.BOUND) ||
                (window.variables[idx].type ==
                 Components.interfaces.mozCellMLIVariable.SIMPLE);
  tf.value = value;
  var lb = document.getElementById("fieldname");
  lb.value = name + ": ";
}

function TypeToString(v)
{
  switch (v)
  {
    case Components.interfaces.mozCellMLIVariable.CONSTANT:
      return "constant";
    case Components.interfaces.mozCellMLIVariable.SIMPLE:
      return "mathematical";
    case Components.interfaces.mozCellMLIVariable.DIFFERENTIAL:
      return "differential";
    case Components.interfaces.mozCellMLIVariable.BOUND:
      return "bound";
  }
}

function AddVariables(box)
{
  window.model =
  window.arguments[0].QueryInterface(Components.interfaces.mozICellML);

  var statusBar = document.getElementById('status-bar-display');
  var validity = window.model.checkValidity();
  if (validity == "")
    statusBar.label = "Model loaded with no errors";
  else
    statusBar.label = "Model is invalid: " + validity;

  window.integrator = window.arguments[1];
  var varit = window.model.getAllVariables(Components.interfaces.
                                           mozICellML.NO_INTERFACE_IN);
  window.variables = new Array();
  var i = 0;
  while (varit.hasMoreElements())
  {
    var v = varit.getNext();
    v = v.QueryInterface(Components.interfaces.mozCellMLIVariable);
    if (v.type == Components.interfaces.mozCellMLIVariable.BOUND)
    {
      window.bound = v.name;
      window.boundunit = v.units;
      var boundname = document.getElementById("boundname");
      boundname.value = window.bound + " interval: (";
      var unitsep = document.getElementById("unitsep");
      unitsep.value = window.boundunit + " ,";
      var unitstop = document.getElementById("unitstop");
      unitstop.value = window.boundunit + ")";
      var unitgrid = document.getElementById("unitgrid");
      unitgrid.value = window.boundunit;
    }
    window.variables[i] = v;
    var iv = "";
    try
    {
      iv = v.initialValue;
    } catch(e) {}
    InsertVariable(box, v.name, TypeToString(v.type), iv, v.units);
    i++;
  }
  window.nvariables = i;

  if (window.integrator != null)
  {
    // XXX we need to synchronise whether an integration is in
    // progress between all the windows that share integrators.
    ChangeIntegratorButtons(document, true, false);
  }
}

function SelectLocalSave(returnto)
{
  try
  {
    const nsIFilePicker = Components.interfaces.nsIFilePicker;
    var fp = Components.classes["@mozilla.org/filepicker;1"]
                       .createInstance(nsIFilePicker);
    fp.init(window, "Choose filename to save to",
            nsIFilePicker.modeSave);
    fp.appendFilter("All CellML", "*.cml; *.xml");
    fp.appendFilter("CML", "*.cml");
    fp.appendFilter("XML", "*.xml");
    fp.appendFilter("All Files", "*");
    if (fp.show() == nsIFilePicker.returnCancel)
      return;
    returnto.value = fp.file.path;
  }
  catch (ex)
  {
  }
}

function SelectLocalExport(returnto)
{
  try
  {
    const nsIFilePicker = Components.interfaces.nsIFilePicker;
    var fp = Components.classes["@mozilla.org/filepicker;1"]
                       .createInstance(nsIFilePicker);
    fp.init(window, "Select CSV file export location",
            nsIFilePicker.modeSave);
    fp.appendFilter("CSV", "*.csv");
    fp.appendFilter("Text Files", "*.txt");
    fp.appendFilter("All Files", "*");
    if (fp.show() == nsIFilePicker.returnCancel)
      return;
    returnto.value = fp.file.path;
  }
  catch (ex)
  {
  }
}

function UserRequestSave(saveto, model)
{
  var file = Components.classes["@mozilla.org/file/local;1"]
                    .createInstance(Components.interfaces.nsILocalFile);
  try
  {
    file.initWithPath(saveto);
  }
  catch (e)
  {
    var statusBar = document.getElementById('status-bar-display');
    statusBar.label = "Can't save to that path.";
    return;
  }
  var stream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                    .createInstance(Components.interfaces.nsIFileOutputStream);
  var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"]
                    .createInstance(Components.interfaces.nsIDOMSerializer);
  stream.init(
              file,
              /* PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE */
              42,
              /*0644*/420,
              0
             );
  var savedom = model.dom;
  serializer.serializeToStream(savedom, stream, "UTF8");
}

function SetVisibility(x, isvis)
{
  x.setAttribute("style", "visibility: " + (isvis ? "visible" : "collapse"));
}

function ChangeIntegratorButtons(d, intg, run)
{
  var expert = 0;
  try
  {
    var prefservice =
      Components.classes["@mozilla.org/preferences-service;1"]
               .getService(Components.interfaces.nsIPrefService);
    var cellBranch = prefservice.getBranch("physiome.mozcellml.");
    expert = cellBranch.getIntPref("expertlevel");
  }
  catch (e) {}

  SetVisibility(d.getElementById("compile-integrator"), !intg);
  SetVisibility(d.getElementById("kill-integrator"), intg && (expert > 1));
  SetVisibility(d.getElementById("start-integration"), (!run) && intg);
  SetVisibility(d.getElementById("cancel-integration"), run && intg);
}

function CompileIntegrator()
{
  // Get a temporary file name, to store the C code in...
  var dirService = Components.classes["@mozilla.org/file/directory_service;1"]
                             .getService(Components
                                         .interfaces
                                         .nsIDirectoryServiceProvider);
  var p = new Object();
  var lf = dirService.getFile("TmpD", p)
                     .QueryInterface(Components.interfaces.nsILocalFile);
  lf.appendRelativePath("integrator.c");
  lf.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
                  /*0600*/384);
  var cfile = lf.path;
  var lfx = dirService.getFile("TmpD", p)
                      .QueryInterface(Components.interfaces.nsILocalFile);
  if (navigator.platform.substring(0, 5) == "Win32")
    lfx.appendRelativePath("integrator.exe");
  else
    lfx.appendRelativePath("integrator");

  lfx.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
                  /*0700*/448);
  var xfile = lfx.path;

  var cc = GetCompiler();

  // XXX not all compilers take the same arguments...
  cc += " " + cfile + " -o " + xfile;

  var statusBar = document.getElementById('status-bar-display');

  statusBar.label = "Compiling the integrator...";
  ChangeIntegratorButtons(document, true, false);
  try
  {
    window.integrator = window.model.newIntegrator(cfile, cc, xfile);
    statusBar.label = "Integrator compiled.";
  }
  catch (e)
  {
    var valerr = window.model.checkValidity();
    if (valerr == "")
      statusBar.label = "Integration failed.";
    else
    {
      statusBar.label = "Integration failed, model is invalid: " + valerr;
    }
    ChangeIntegratorButtons(document, false, false);
  }
  lf.remove(false);
}

function CleanupWindow()
{
  if (window.integrator != null)
  {
    window.integrator.destroy();
    window.integrator = null;
  }
}

function KillIntegrator()
{
  var statusBar = document.getElementById('status-bar-display');
  statusBar.label = "Killing numerical integrator...";
  window.integrator.destroy();
  window.integrator = null;
  ChangeIntegratorButtons(document, false, false);
  statusBar.label = "Numerical integrator killed.";
}

function IntegrationResultWatcher(npoints)
{
  this.data = new Array();
  this.curpoint = 0;
  this.npoints = npoints + 1;
  return this;
}

IntegrationResultWatcher.prototype =
{
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.mozCellMLIPointWatcher) ||
        aIID.equals(Components.interfaces.nsISupports) ||
        aIID.equals(Components.interfaces.nsISupportsWeakReference)
       )
      return this;
    throw Components.results.NS_NOINTERFACE;
  },

  pointAdded: function(x, numvariables, variables, values)
  {
    var i, j;
    if (this.varmap == null)
    {
      this.varmap = new Array();
      for (i = 0; i < numvariables; i++)
      {
        for (j = 0; j < this.nvariables; j++)
          if (this.variables[j].equals(variables[i]))
          {
            this.varmap[i] = j;
            break;
          }
      }
    }
    this.data[this.curpoint] = new Array();
    for (i = 0; i < numvariables; i++)
      this.data[this.curpoint][this.varmap[i]] = values[i];
    this.curpoint++;
    if ((this.curpoint & 0xF) == 0)
    {
      if (this.meter != null)
        this.meter.setAttribute("value",
                                ((this.curpoint / this.npoints) * 100));
    }
  },

  done: function()
  {
    var statusBar = document.getElementById('status-bar-display');
    statusBar.label = "Numerical Integration Complete";
    if (window.integrator != null)
      ChangeIntegratorButtons(document, true, false);
    this.curpoint = this.npoints;
    this.live = false;
    if (this.meter != null)
      this.meter.setAttribute("value", "100%");
  },

  halted: function()
  {
    var statusBar = document.getElementById('status-bar-display');
    statusBar.label = "Numerical Integration Halted";
    this.live = false;
    if (window.integrator != null)
      ChangeIntegratorButtons(document, true, false);
  },

  failed: function(emsg)
  {
    var statusBar = document.getElementById('status-bar-display');
    statusBar.label = "Numerical Integration Failed: " + emsg;
    alert("Numerical Integration Failed: " + emsg);
    this.live = false;
    if (window.integrator != null)
      ChangeIntegratorButtons(document, true, false);
  },

  live: false,
  closed: false
};

function StartIntegration()
{
  var statusBar = document.getElementById('status-bar-display');
  var startx = parseFloat(document.getElementById('startx').value);
  var stopx = parseFloat(document.getElementById('stopx').value);
  var gridx = parseFloat(document.getElementById('gridx').value);
  var relerr = parseFloat(document.getElementById('relerr').value);
  var abserr = parseFloat(document.getElementById('relerr').value);
  if (isNaN(startx) || !isFinite(startx))
  {
    statusBar.label = "Invalid starting X value.";
    return;
  }
  if (isNaN(stopx) || !isFinite(stopx))
  {
    statusBar.label = "Invalid ending X value.";
    return;
  }
  if (isNaN(gridx) || !isFinite(gridx))
  {
    statusBar.label = "Invalid X resolution value.";
    return;
  }
  if (isNaN(relerr) || !isFinite(relerr))
  {
    statusBar.label = "Invalid relative error value.";
    return;
  }
  if (isNaN(abserr) || !isFinite(abserr))
  {
    statusBar.label = "Invalid absolute error value.";
    return;
  }
  var diff = stopx - startx;
  if (diff < 0)
    diff = -diff;
  var irw = new IntegrationResultWatcher(parseInt(diff / gridx));
  irw.startx = startx;
  irw.stopx = stopx;
  irw.gridx = gridx;
  irw.model = window.model;
  irw.variables = new Array();
  var i = 0, j = 0;
  for (i = 0; i < window.nvariables; i++)
  {
    if (window.variables[i].type !=
        Components.interfaces.mozCellMLIVariable.BOUND &&
        window.variables[i].type !=
        Components.interfaces.mozCellMLIVariable.CONSTANT)
      irw.variables[j++] = window.variables[i];
  }
  irw.nvariables = j;
  irw.integrator = window.integrator;
  irw.maindoc = document;
  irw.mainwin = window;
  irw.bound = window.bound;
  irw.boundunit = window.boundunit;
  window.integrator.registerWatcher(irw);
  window.openDialog(
                    "chrome://mozcellml/content/integrationresult.xul",
                    "_blank",
                    "chrome,dialog=no",
                    irw
                   );
  irw.live = true;
  if (!irw.closed)
  {
    window.integrator.start(window.model, startx, stopx, gridx,
                            abserr, relerr);
    ChangeIntegratorButtons(document, true, true);
    statusBar.label = "Integration Started";
  }
}

function CancelIntegrationWindow()
{
  window.arguments[0].closed = true;
  if (!window.arguments[0].live)
    return;
  if (window.arguments[0].mainwin.integrator != null)
  {
    ChangeIntegratorButtons(window.arguments[0].maindoc, true, false);
    window.arguments[0].mainwin.integrator.suspend();
  }
}

function CancelIntegration()
{
  if (window.integrator != null)
  {
    ChangeIntegratorButtons(document, true, false);
    window.integrator.suspend();
  }
}

function nyble2hex(x)
{
  if (x < 10)
    return "" + x;
  switch (x)
  {
    case 10:
      return 'A';
    case 11:
      return 'B';
    case 12:
      return 'C';
    case 13:
      return 'D';
    case 14:
      return 'E';
    case 15:
      return 'F';
  }
}

function hex2nyble(x)
{
  var c = x.charCodeAt(0);
  if (c >= 0x30 && c <= 0x39)
    return c - 0x30;
  if (c >= 0x41 && c <= 0x46)
    return c - 0x37;
  return c - 0x57;
}

function ascolour(val)
{
  return "#" + nyble2hex((val & 0xf00000)>>20) +
               nyble2hex((val & 0xf0000)>>16) +
               nyble2hex((val & 0xf000)>>12) +
               nyble2hex((val & 0xf00)>>8) +
               nyble2hex((val & 0xf0)>>4) +
               nyble2hex(val & 0xf);
}

function colourtonumber(col)
{
  return (hex2nyble(col.substring(1, 2)) << 20) |
         (hex2nyble(col.substring(2, 3)) << 16) |
         (hex2nyble(col.substring(3, 4)) << 12) |
         (hex2nyble(col.substring(4, 5)) << 8) |
         (hex2nyble(col.substring(5, 6)) << 4) |
         hex2nyble(col.substring(6, 7));
}

function InsertIntegrationVariable(box, name, type, colour)
{
  var li = document.createElement("listitem");
  li.setAttribute("id", name);
  var lc = document.createElement("listcell");
  lc.setAttribute("label", name);
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("label", type);
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("label", "Yes");
  li.appendChild(lc);
  lc = document.createElement("listcell");
  lc.setAttribute("style", "background-color: " + ascolour(colour));
  li.appendChild(lc);
  box.appendChild(li);
}

function PrepareIntegrationWindow()
{
  var boundname = document.getElementById("boundname");
  boundname.value = window.arguments[0].bound;
  var boundunit = document.getElementById("boundunit");
  boundunit.value = window.arguments[0].boundunit;

  window.arguments[0].meter = document.getElementById("integration-progress");
  window.arguments[0].meter.setAttribute("value",
      ((window.arguments[0].curpoint / window.arguments[0].npoints) * 100));

  var box = document.getElementById("variable-tree");
  window.model = window.arguments[0].model;
  window.variables = window.arguments[0].variables;
  window.nvariables = window.arguments[0].nvariables;
  window.show = new Array();

  var slider = document.getElementById("xpos-scroll");
  var xbox = document.getElementById("xpos");
  slider.setAttribute("maxpos",
                      (window.arguments[0].stopx - window.arguments[0].startx) / 
                      window.arguments[0].gridx);
  xbox.value = window.arguments[0].startx;
  var i = 0;
  var nv = window.nvariables;
  var col = new Array();

  for (i = 0; i < nv; i++)
  {
    window.show[i] = true;
    var v = window.variables[i];
    col[i] = Math.random() * 0xffffff;
    InsertIntegrationVariable(box, v.name, TypeToString(v.type), col[i]);
  }
  window.colours = col;

  window.arguments[0].intwindow = window;
}

function GetXAtTime(ti)
{
  var statusBar = document.getElementById("status-bar-display");
  if (ti < 0)
  {
    statusBar.label = "That is before the first data point.";
    return;
  }
  if (ti >= window.arguments[0].curpoint)
  {
    statusBar.label = "No data is available for that time.";
    return;
  }
  var data = window.arguments[0].data[ti];
  if (data == null)
  {
    statusBar.label = "No data is available for that time.";
    return;
  }
  var box = document.getElementById("variable-tree");
  var idx = box.selectedIndex;
  if (idx == -1)
  {
    statusBar.label = "Select a variable first.";
    return;
  }
  document.getElementById("value-at-x").value = data[idx];
  statusBar.label = "X Value retrieved";
}

function SelectIntegrationVariable()
{
  var box = document.getElementById("variable-tree");
  var idx = box.selectedIndex;
  if (idx < 0)
    return;
  var name = window.variables[idx].name;
  document.getElementById("variable-name").value = name;
  document.getElementById("variable-unit").value = window.variables[idx].units;
  document.getElementById("draw-on-graph").checked = window.show[idx];
  var col = ascolour(window.colours[idx]);
  document.getElementById("graph-colour").color = col;
  GetXAtTime(document.getElementById("xpos-scroll").getAttribute("curpos"));
  return true;
}

function ChangeGraphColour()
{
  var box = document.getElementById("variable-tree");
  var idx = box.selectedIndex;
  if (idx < 0)
    return;
  var colour = document.getElementById("graph-colour").color;
  window.colours[idx] = colourtonumber(colour);
  box.selectedItem.childNodes.item(3)
     .setAttribute("style", "background-color: " + colour);
  return true;
}

function ChangeDraw()
{
  var box = document.getElementById("variable-tree");
  var idx = box.selectedIndex;
  if (idx < 0)
    return;
  var draw = document.getElementById("draw-on-graph").checked;
  window.show[idx] = draw;
  box.selectedItem.childNodes.item(2)
     .setAttribute("label", draw ? "Yes" : "No");
  return false;
}

function OnXSlide()
{
  var slider = document.getElementById("xpos-scroll");
  var box = document.getElementById("xpos");
  box.value = slider.getAttribute("curpos") *
              window.arguments[0].gridx +
              window.arguments[0].startx;
  GetXAtTime(slider.getAttribute("curpos"));
  return true;
}

function OnXTextChange()
{
  var slider = document.getElementById("xpos-scroll");
  var box = document.getElementById("xpos");
  slider.setAttribute("curpos",
                      Math.floor((box.value - window.arguments[0].startx) /
                                 window.arguments[0].gridx));
  GetXAtTime(slider.getAttribute("curpos"));
  return true;
}

function CloneAtTime()
{
  var ti = document.getElementById("xpos-scroll").getAttribute("curpos");
  var statusBar = document.getElementById("status-bar-display");
  if (ti < 0)
  {
    statusBar.label = "That is before the first data point.";
    return;
  }
  if (ti >= window.arguments[0].curpoint)
  {
    statusBar.label = "No data is available for that time.";
    return;
  }
  var data = window.arguments[0].data[ti];
  var olddom = window.model.dom;
  var newdom = olddom.cloneNode(true);
  newdom = newdom.QueryInterface(Components.interfaces.nsIDOMDocument);
  statusBar.label = "Copying model structure...";
  var newmodel = Components.classes["@physiome.org/mozCellML;1"]
                  .createInstance(Components.interfaces.mozICellML);
  newmodel.dom = newdom;
  newmodel.initialise();
  var i;
  statusBar.label = "Copying variables into new model...";
  for (i = 0; i < window.nvariables; i++)
  {
    try
    {
      // Throw an exception if no initial value...
      var x = window.variables[i].initialValue;
      newmodel.getVariable(window.variables[i].name).initialValue =
        data[i];
    }
    catch (e)
    {
      // If there was no initial value before, we don't want one now.
    }
  }
  statusBar.label = "Creating new model window...";
  window.arguments[0].integrator.want();
  window.openDialog(
                    "chrome://mozcellml/content/modelview.xul",
                    "_blank",
                    "chrome,dialog=no",
                    newmodel,
                    window.arguments[0].integrator
                   );
  statusBar.label = "Model cloned";
}

function StreamPrint(strm, strng)
{
  strng += "\r\n";
  strm.write(strng, strng.length);
}

function GetVariableDescriptor(v, i)
{
  var vd = v[i].name;
  try
  {
    vd += "(" + v[i].units + ")";
  }
  catch (e)
  {
  }
  return vd;
}

function SaveCSV()
{
  var filename = document.getElementById('csvfile').value;
  var file = Components.classes["@mozilla.org/file/local;1"]
                       .createInstance(Components.interfaces.nsILocalFile);
  try
  {
    file.initWithPath(filename);
  }
  catch (e)
  {
    var statusBar = document.getElementById('status-bar-display');
    statusBar.label = "Can't save to that path.";
    return;
  }
  var stream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                    .createInstance(Components.interfaces.nsIFileOutputStream);
  stream.init(
              file,
              /* PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE */
              42,
              /*0644*/420,
              0
             );
  var intg = window.arguments[0];
  var names = "";
  if (intg.nvariables == 0)
  {
    stream.close();
    return;
  }
  names += intg.bound + "(" + intg.boundunit + ")";
  for (i = 0; i < intg.nvariables; i++)
    names += "," + GetVariableDescriptor(intg.variables, i);
  StreamPrint(stream, names);
  var boundval = intg.startx;
  for (i = 0; i < intg.curpoint; i++)
  {
    var values = boundval;
    boundval += intg.gridx;
    for (j = 0; j < intg.nvariables; j++)
      values += "," + intg.data[i][j];
    StreamPrint(stream, values);
  }
  stream.close();
  var statusBar = document.getElementById('status-bar-display');
  statusBar.label = "CSV Export Complete.";
}
